﻿// Copyright (c) 2020 Tudor Stanciu

using System;
using System.Threading.Tasks;
using Tuitio.Application.Services.Abstractions;
using Tuitio.Application.Stores;
using Tuitio.Domain.Abstractions;
using Tuitio.Domain.Entities;
using Tuitio.Domain.Models;
using Tuitio.Domain.Models.Account;
using Tuitio.Domain.Repositories;

namespace Tuitio.Application.Services
{
    internal class UserService : IUserService
    {
        private readonly ITokenStore _securityStore;
        private readonly IUserRepository _userRepository;
        private readonly ITokenService _tokenService;
        private readonly IConfigProvider _configProvider;
        private readonly IHashingService _hashingService;

        public UserService(ITokenStore securityStore, IUserRepository userRepository, ITokenService tokenService, IConfigProvider configProvider, IHashingService hashingService)
        {
            _securityStore=securityStore;
            _userRepository=userRepository;
            _tokenService=tokenService;
            _configProvider=configProvider;
            _hashingService=hashingService;
        }

        public async Task<LoginResult> Login(string userName, string password)
        {
            ValidateCredentials(userName, password);

            var passwordHash = _hashingService.HashSha256(password);
            var user = await _userRepository.GetUser(userName, passwordHash);
            if (user == null)
                return null;

            var valid = ValidateUser(user);
            if (!valid)
                return null;

            var tokenExtension = _tokenService.GenerateToken(user);
            var raw = tokenExtension.Token.Export();
            _securityStore.Set(raw, tokenExtension);

            await _userRepository.UpdateUserAfterLogin(user, tokenExtension, raw);

            var result = new LoginResult(tokenExtension, raw);
            return result;
        }

        public async Task<LogoutResult> Logout(string tokenRaw)
        {
            var tokenExtension = _securityStore.Get(tokenRaw);
            if (tokenExtension == null)
                return null;

            await _userRepository.RemoveToken(tokenExtension.Token.TokenId);
            _securityStore.Remove(tokenRaw);

            var result = new LogoutResult(tokenExtension.Token.UserId, tokenExtension.Token.UserName, DateTime.UtcNow);
            return result;
        }

        public TokenExtension Authorize(string tokenRaw)
        {
            var token = _securityStore.Get(tokenRaw);
            return token;
        }

        private void ValidateCredentials(string userName, string password)
        {
            if (string.IsNullOrEmpty(userName))
                throw new ArgumentException($"Value cannot be null or empty string.", nameof(userName));

            if (string.IsNullOrEmpty(password))
                throw new ArgumentException($"Value cannot be null or empty string.", nameof(password));
        }

        private bool ValidateUser(AppUser user)
        {
            if (user == null)
                throw new ArgumentNullException(nameof(user));

            if (user.FailedLoginAttempts.HasValue && user.FailedLoginAttempts.Value > _configProvider.Restrictions.MaxFailedLoginAttempts)
                return false;

            return true;
        }
    }
}
