﻿// Copyright (c) 2020 Tudor Stanciu

using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using Tuitio.Domain.Helpers;

namespace Tuitio.Domain.Models
{
    public class Token
    {
        public Guid TokenId { get; set; }
        public int UserId { get; set; }
        public string UserName { get; set; }
        public string FirstName { get; set; }
        public string LastName { get; set; }
        public string Email { get; set; }
        public string SecurityStamp { get; set; }
        public string LockStamp { get; set; }
        public DateTime CreatedAt { get; set; }
        public long ExpiresIn { get; set; }
        public Dictionary<string, string> Claims { get; set; }

        [Obsolete("This constructor is only used for deserialization and should not be used for any other purpose.")]
        public Token() { }

        private Token(int validityInMinutes)
        {
            TokenId = Guid.NewGuid();
            CreatedAt = DateTime.UtcNow;
            LockStamp = Regex.Replace(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), "[/+=]", "");
            ExpiresIn = validityInMinutes * 60; // seconds
        }

        public static Token Initialize(int validityInMinutes, int userId, string userName, string firstName, string lastName, string email, string securityStamp, Dictionary<string, string> claims)
        {
            var token = new Token(validityInMinutes)
            {
                UserId = userId,
                UserName = userName,
                FirstName = firstName,
                LastName = lastName,
                Email = email,
                SecurityStamp = securityStamp,
                Claims = claims
            };

            return token;
        }

        public string Export()
        {
            var tokenString = JsonConvert.SerializeObject(this);
            var tokenBytes = Encoding.UTF8.GetBytes(tokenString);
            var tokenRaw = Convert.ToBase64String(tokenBytes);
            return tokenRaw;
        }

        public static Token Import(string tokenRaw)
        {
            var valid = ValidateTokenRaw(tokenRaw);
            if (!valid)
                return null;

            var tokenBytes = Convert.FromBase64String(tokenRaw);
            var tokenString = Encoding.UTF8.GetString(tokenBytes);
            var token = JsonConvert.DeserializeObject<Token>(tokenString);
            return token;
        }

        public TokenExtension Extend(IEnumerable<RecordIdentifier> userRoles, IEnumerable<RecordIdentifier> userGroups)
            => TokenExtension.Create(this, userRoles, userGroups);

        private static bool ValidateTokenRaw(string tokenRaw)
        {
            if (string.IsNullOrWhiteSpace(tokenRaw))
                return false;

            if (!DataValidationHelper.StringIsBase64(tokenRaw))
                return false;

            return true;
        }
    }
}
